import os
import re
import math
import numpy as np
import cv2
import warnings
warnings.filterwarnings("ignore", category=FutureWarning)

import gst_helper
import gst_cv_helper
import gst_admeta as admeta
from gi.repository import Gst, GObject, GLib, GstVideo

import sys
import random
sys.path.insert(1, "/opt/adlink/eva/plugins/python/leda_object_detection")
from leda_object_detection.models.experimental import attempt_load
from leda_object_detection.utils.torch_utils import select_device
from leda_object_detection.utils.datasets import LoadImages

import time
from pathlib import Path

import cv2
import torch
import torch.backends.cudnn as cudnn
from numpy import random

sys.path.insert(1, "/opt/adlink/eva/plugins/python/leda_object_detection")
from leda_object_detection.models.experimental import attempt_load
from leda_object_detection.utils.datasets import LoadStreams, LoadImages
from leda_object_detection.utils.general import check_img_size, check_requirements, check_imshow, non_max_suppression, apply_classifier, \
    scale_coords, xyxy2xywh, strip_optimizer, set_logging, increment_path
from leda_object_detection.utils.plots import plot_one_box
from leda_object_detection.utils.torch_utils import select_device, load_classifier, time_synchronized

##### Analyze inference result context #####

def parse_results_leda(results, image_size, padding, threshold=0.8):
	if results is None or len(results) != 3:
		return []
	bboxes, categories, confidences = results
	if bboxes is None or len(bboxes) <= 0 or len(bboxes) != len(categories) or len(bboxes) != len(confidences):
		return []
	ih, iw = image_size
	parsed_results = []
	for i, (box, category, confidence) in enumerate(zip(bboxes, categories, confidences)):
		if confidence < threshold:
			continue
		x_coord, y_coord, x2, y2 = box
		left = x_coord
		top = y_coord
		right = x2-x_coord
		bottom = y2-y_coord
		parsed_results.append((i, category, left / iw, top / ih, right / iw, bottom / ih, confidence))
	return parsed_results

def parse_results(results, image_size, padding, threshold=0.8):
	if results is None or len(results) != 3:
		return []
	bboxes, categories, confidences = results
	if bboxes is None or len(bboxes) <= 0 or len(bboxes) != len(categories) or len(bboxes) != len(confidences):
		return []
	ih, iw = image_size
	parsed_results = []
	for i, (box, category, confidence) in enumerate(zip(bboxes, categories, confidences)):
		if confidence < threshold:
			continue
		x_coord, y_coord, width, height = box
		left = max(0, np.floor(x_coord - padding * iw).astype(int))
		top = max(0, np.floor(y_coord - padding * ih).astype(int))
		right = min(iw, np.floor(x_coord + width + padding * iw).astype(int))
		bottom = min(ih, np.floor(y_coord + height + padding * ih).astype(int))
		parsed_results.append((i, category, left / iw, top / ih, right / iw, bottom / ih, confidence))
	return parsed_results

############################################

def letterbox(img, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, stride=32):
    # Resize and pad image while meeting stride-multiple constraints
    shape = img.shape[:2]  # current shape [height, width]
		
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2
    	
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return img, ratio, (dw, dh)


##### GStreamer python plugin context #####

def gst_video_caps_make(fmt):
	return "video/x-raw, "\
		"format = (string) " + fmt + " , "\
		"width = " + GstVideo.VIDEO_SIZE_RANGE + ", "\
		"height = " + GstVideo.VIDEO_SIZE_RANGE + ", "\
		"framerate = " + GstVideo.VIDEO_FPS_RANGE

class LedaDetectorPy(Gst.Element):
	GST_PLUGIN_NAME = 'leda_detector_py'
	__gstmetadata__ = ("LedaDetector", "GstElement", "This is Leda Detector.", "Leda RD")
	__gsttemplates__ = (Gst.PadTemplate.new(
			"src",
			Gst.PadDirection.SRC,
			Gst.PadPresence.ALWAYS,
			Gst.Caps.from_string(gst_video_caps_make("{ BGR }"))
		),
		Gst.PadTemplate.new(
			"sink",
			Gst.PadDirection.SINK,
			Gst.PadPresence.ALWAYS,
			Gst.Caps.from_string(gst_video_caps_make("{ BGR }"))
		)
	)
	_sinkpadtemplate = __gsttemplates__[1]
	_srcpadtemplate = __gsttemplates__[0]
	__gproperties__ = {
#		"name": (data type, name, blurb, value, flags)
#		"name": (int, "name", "blurb", min, max, value, GObject.ParamFlags.READWRITE)
#		"name": (float, "name", "blurb", min, max, value, GObject.ParamFlags.READWRITE)
#		"name": (str, "name", "blurb", value, GObject.ParamFlags.READWRITE)
		"class_num": (int, "class-num", "Class number", 1, 65536, 80, GObject.ParamFlags.READWRITE),
		"batch_num": (int, "batch-num", "Batch number", 1, 256, 1, GObject.ParamFlags.READWRITE),
		"padding": (float, "padding", "YOLO padding", 0.0, 1.0, 0.0, GObject.ParamFlags.READWRITE),
		"threshold": (float, "threshold", "Threshold of box confidence for yolo", 0.0, 1.0, 0.8, GObject.ParamFlags.READWRITE),
		"cthreshold": (float, "cthreshold", "Threshold of box confidence for addrawer", 0.0, 1.0, 0.8, GObject.ParamFlags.READWRITE),
		"label_file": (str, "label-file", "Label file", "", GObject.ParamFlags.READWRITE),
		"model_path": (str, "model-path", "Test read and output file", "this/is/test", GObject.ParamFlags.READWRITE),
	}

	def __init__(self):
		print(">" * 20, "Initial start.")
		super(LedaDetectorPy, self).__init__()
		self.device = select_device('')
		self.model = None #attempt_load(get_defeault_weights(), map_location=device)
		self.stride = 0
		self.imgsz = 640
		self.class_num = 80
		self.batch_num = 1
		self.padding = 0
		self.threshold = 0.3
		self.cthreshold = 0.3
		self.label_file = "model/yolo_RT_labels.txt"
		if os.path.isfile(self.label_file):
			with open(self.label_file, 'r') as f:
				self.labels = list(map(lambda l: l.strip(), f))
		else:
			self.labels = []
		self.model_path = "this/is/test"
		
		self.sinkpad = Gst.Pad.new_from_template(self._sinkpadtemplate, 'sink')
		self.sinkpad.set_chain_function_full(self.chainfunc, None)
		self.sinkpad.set_chain_list_function_full(self.chainlistfunc, None)
		self.sinkpad.set_event_function_full(self.eventfunc, None)
		self.add_pad(self.sinkpad)
		self.srcpad = Gst.Pad.new_from_template(self._srcpadtemplate, 'src')
		self.srcpad.set_event_function_full(self.srceventfunc, None)
		self.srcpad.set_query_function_full(self.srcqueryfunc, None)
		self.add_pad(self.srcpad)
		print(">" * 20, "Initial done.")

	def do_get_property(self, prop: GObject.GParamSpec):
		if prop.name == 'class-num':
			return self.class_num
		elif prop.name == 'batch-num':
			return self.batch_num
		elif prop.name == 'padding':
			return self.padding
		elif prop.name == 'threshold':
			return self.threshold
		elif prop.name == 'cthreshold':
			return self.cthreshold
		elif prop.name == 'label-file':
			return self.label_file
		elif prop.name == 'model-path':
			return self.model_path
		else:
			raise AttributeError('unknown property %s' % prop.name)
		#return getattr(self, prop.name)

	def do_set_property(self, prop: GObject.GParamSpec, value):
		if prop.name == 'class-num':
			self.class_num = int(value)
		elif prop.name == 'batch-num':
			self.batch_num = int(value)
		elif prop.name == 'padding':
			self.padding = float(value)
		elif prop.name == 'threshold':
			self.threshold = float(value)
		elif prop.name == 'cthreshold':
			self.cthreshold = float(value)
		elif prop.name == 'label-file':
			self.label_file = str(value)
			if os.path.isfile(self.label_file):
				with open(self.label_file, 'r') as f:
					self.labels = list(map(lambda l: l.strip(), f))
		elif prop.name == 'model-path':
			self.model_path = str(value)
			print(self.model_path)
			device = select_device('')
			with torch.no_grad():			
				self.model = attempt_load(self.model_path, map_location=device)
				self.stride = int(self.model.stride.max())  # model stride
				self.imgsz = 640  # check img_size
				self.model.half()  # to FP16
		else:
			raise AttributeError('unknown property %s' % prop.name)

	def chainfunc(self, pad: Gst.Pad, parent, buff: Gst.Buffer) -> Gst.FlowReturn:
		#print(">" * 20, "chainfunc() start.")
		
		leda_detector_args = {
			"model_path": self.model_path,
		}
		self.labels = self.model.names
		img = gst_cv_helper.pad_and_buffer_to_numpy(pad, buff, ro=False)
		h, w, c = img.shape
		im0 = img.copy()
		img = img.astype(np.float32)
		img = letterbox(img, (640,640), stride=self.stride)[0]
		img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
		img = np.ascontiguousarray(img)
		img = torch.from_numpy(img).to(self.device)
		img = img.half()
		img /= 255.0  # 0 - 255 to 0.0 - 1.0
		if img.ndimension() == 3:
			img = img.unsqueeze(0)

		# Inference
		pred = self.model(img, augment=False)[0]		
		# Apply NMS
		pred = non_max_suppression(pred, self.threshold, 0.45, classes=None, agnostic=False)
		boxes = []
		classes = []
		scores = []
		_boxes = []		
		for i, det in enumerate(pred): 
			if len(det):
				det[:, :4] = scale_coords(img.shape[2:], det[:, :4], im0.shape).round()
				for *xyxy, conf, cls in reversed(det):
					b_bbox = torch.tensor(xyxy).tolist()
					B = [int(b_bbox[0]), int(b_bbox[1]),int(b_bbox[0]+b_bbox[2]),int(b_bbox[1]+b_bbox[3])]
					
					boxes.append(B)
					classes.append(int(torch.tensor(cls)))
					scores.append(float(torch.tensor(conf)))
	
		_boxes = parse_results_leda((boxes, classes, scores), (h, w), self.padding, self.cthreshold)
			
		arr = []
		
		for b in _boxes:
			l = self.labels[b[1]] if b[1] < len(self.labels) else str(b[1])
			arr.append(admeta._DetectionBox(b[0], l, b[1] ," ", b[2], b[3], b[4], b[5], b[6], ""))
		admeta.set_detection_box(buff, pad, arr)
		
		#print(">" * 20, "chainfunc() done.")
		return self.srcpad.push(buff)

	def chainlistfunc(self, pad: Gst.Pad, parent, buff_list: Gst.BufferList) -> Gst.FlowReturn:
		print(">" * 20, "chainlistfunc() start.")
		print(">" * 20, "chainlistfunc() done.")
		return self.srcpad.push(buff_list.get(0))

	def eventfunc(self, pad: Gst.Pad, parent, event: Gst.Event) -> bool:
		return self.srcpad.push_event(event)

	def srcqueryfunc(self, pad: Gst.Pad, parent, query: Gst.Query) -> bool:
		return self.sinkpad.query(query)

	def srceventfunc(self, pad: Gst.Pad, parent, event: Gst.Event) -> bool:
		return self.sinkpad.push_event(event)

GObject.type_register(LedaDetectorPy)
__gstelementfactory__ = (LedaDetectorPy.GST_PLUGIN_NAME, Gst.Rank.NONE, LedaDetectorPy)

###########################################
